import os
import re
import numpy as np
import tifffile as tiff
import zarr
import matplotlib.pyplot as plt
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
from functools import partial

def collect_grid_coordinates(base_dir):
    pattern = re.compile(r"cell_yxz_(\d{3})_(\d{3})_(\d{3})\.tif")
    grid_coords = [(int(m.group(1)), int(m.group(2)), int(m.group(3))) 
                   for m in (pattern.match(f) for f in os.listdir(base_dir)) 
                   if m is not None]
    return sorted(grid_coords)

def read_tiff_to_array(filename):
    try:
        # First try memory mapping
        return tiff.memmap(filename, mode='r')
    except (ValueError, Exception):
        return tiff.imread(filename)

def determine_zarr_shape(grid_coords, full_volume_shape=None, grid_size=500):
    if not grid_coords:
        raise ValueError("No grid coordinates found")
        
    if full_volume_shape is not None:
        return full_volume_shape
    else:
        # Calculate based on grid coordinates (using 0-based indexing for calculation)
        max_jy = max(jy for jy, _, _ in grid_coords)
        max_jx = max(jx for _, jx, _ in grid_coords)
        max_jz = max(jz for _, _, jz in grid_coords)
        # Since coordinates are 1-based, multiply (max_j - 1 + 1) * grid_size to get total size
        return ((max_jz) * grid_size, (max_jy) * grid_size, (max_jx) * grid_size)

def process_single_tiff(args):
    try:
        zarr_array, base_dir, coords, grid_size = args
        jy, jx, jz = coords  
        filename = os.path.join(base_dir, f"cell_yxz_{jy:03d}_{jx:03d}_{jz:03d}.tif")
        
        if not os.path.exists(filename):
            print(f"Warning: File not found: {filename}")
            return
            
       
        tiff_data = read_tiff_to_array(filename)
        
        # Convert 1-based filename indices to 0-based array indices
        z_start = (jz - 1) * grid_size  # jz=1 → start at 0
        y_start = (jy - 1) * grid_size  # jy=1 → start at 0
        x_start = (jx - 1) * grid_size  # jx=1 → start at 0
        
        # Verify block won't exceed array bounds
        if (z_start + grid_size > zarr_array.shape[0] or
            y_start + grid_size > zarr_array.shape[1] or
            x_start + grid_size > zarr_array.shape[2]):
            print(f"Warning: Block {filename} exceeds array bounds at position ({z_start}, {y_start}, {x_start})")
            return
        
        # Write the entire chunk at once
        zarr_array[z_start:z_start + grid_size,
                  y_start:y_start + grid_size,
                  x_start:x_start + grid_size] = tiff_data
        
    except Exception as e:
        print(f"Error processing file {filename}: {str(e)}")
        raise

def main(base_dir, zarr_store_path, full_volume_shape=None, max_workers=4):
    grid_size = 500
    chunk_size = grid_size  # Match chunk size to grid size
    
    print("Collecting grid coordinates...")
    grid_coords = collect_grid_coordinates(base_dir)
    if not grid_coords:
        raise ValueError("No valid TIFF files found in the input directory")
    
    print("Determining array shape...")
    zarr_shape = determine_zarr_shape(grid_coords, full_volume_shape, grid_size)
    print(f"Creating Zarr array with shape: {zarr_shape}")
    
    zarr_dtype = np.uint8
    
    # Configure Zarr for better write performance
    compressor = zarr.Blosc(cname='zstd', clevel=3, shuffle=zarr.Blosc.SHUFFLE)
    
    print("Creating Zarr array...")
    zarr_store = zarr.DirectoryStore(zarr_store_path)
    zarr_array = zarr.create(
        store=zarr_store,
        shape=zarr_shape,
        dtype=zarr_dtype,
        chunks=(chunk_size, chunk_size, chunk_size),
        compressor=compressor,
        fill_value=0,  # Set default fill value for regions without data
        overwrite=True
    )
    
    process_args = [(zarr_array, base_dir, coords, grid_size) 
                   for coords in grid_coords]
    
    print(f"Processing {len(process_args)} TIFF files in parallel...")
    failed_files = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for args in process_args:
            future = executor.submit(process_single_tiff, args)
            futures.append(future)
        
        for future in tqdm(futures, total=len(futures), desc="Converting TIFF files"):
            try:
                future.result()
            except Exception as e:
                failed_files.append(str(e))
    
    if failed_files:
        print("\nWarning: Some files failed to process:")
        for error in failed_files:
            print(error)
    else:
        print("\nAll files processed successfully!")
    
    print(f"Final Zarr array shape: {zarr_array.shape}")
    return zarr_array


if __name__ == "__main__":
    base_dir = r"E:\scroll1.volpkg\volume_grids\20230205180739"
    zarr_store_path = r"E:\scroll1.volpkg\volume_grids\gp.zarr"
    
    # specify the full volume shape (z, y, x)
    full_volume_shape = (14376, 8096, 7888)
   
    
    max_workers = 10
    
    zarr_array = main(base_dir, zarr_store_path, full_volume_shape=full_volume_shape, max_workers=max_workers)